﻿using Company.Infrastructure.Common;
using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Linq;
using System.Net;
using System.ServiceModel;
using System.ServiceModel.Channels;
using System.Text;
using System.Threading.Tasks;

namespace Company.Infrastructure.Client
{
    public abstract class PolicyProxyBase<I> : IDisposable where I : class
    {
        object channelLock = new object();
        protected ChannelFactory<I> Factory
        { get; set; }

        protected virtual TimeSpan DefaultConnectivityTimeout
        {
            get
            {
                return BindingHelper.DefaultConnectivityTimeout;
            }
        }
        protected virtual TimeSpan DefaultDebugTimeout
        {
            get
            {
                return BindingHelper.DefaultDebugTimeout;
            }
        }
        protected virtual long DefaultMaxMessageSize
        {
            get
            {
                return BindingHelper.DefaultMaxMessageSize;
            }
        }

        // DESIGN NOTE:
        // - Provide the option to initialize separate from instantiation.
        // - Allow for manual overrride of key aspects of your configuration policy
        protected void Initialize(Uri baseAddress)
        {
            Initialize(baseAddress, null, DefaultMaxMessageSize, DefaultConnectivityTimeout);
        }
        protected void Initialize(Uri baseAddress, long maxMessageSize)
        {
            Initialize(baseAddress, null, maxMessageSize, DefaultConnectivityTimeout);
        }
        protected void Initialize(Uri baseAddress, TimeSpan connectivityTimeout)
        {
            Initialize(baseAddress, null, DefaultMaxMessageSize, connectivityTimeout);
        }
        protected void Initialize(Uri baseAddress, NetworkCredential credentials)
        {
            Initialize(baseAddress, credentials, DefaultMaxMessageSize, DefaultConnectivityTimeout);
        }
        protected virtual void Initialize(Uri baseAddress, NetworkCredential credentials, long maxMessageSize, TimeSpan connectivityTimeout)
        {
            ValidateContract();

            Factory = CreateFactory(baseAddress, maxMessageSize, connectivityTimeout);

            if (credentials != null)
            {
                Factory.Credentials.Windows.ClientCredential = credentials;
            }

            Factory.Faulted += Factory_Faulted;
        }

        ChannelFactory<I> CreateFactory(Uri baseAddress, long maxMessageSize, TimeSpan connectivityTimeout)
        {
            if (baseAddress == null)
            {
                throw new ArgumentNullException("baseAddress");
            }

            Binding binding = Binding(maxMessageSize, connectivityTimeout, DefaultDebugTimeout);

            PolicyChannelFactory<I> factory = CreatePolicyChannelFactory(binding, baseAddress);

            return factory;
        }

        protected virtual PolicyChannelFactory<I> CreatePolicyChannelFactory(Binding binding, Uri baseAddress)
        {
            return new PolicyChannelFactory<I>(binding, baseAddress);
        }

        protected virtual void Factory_Faulted(object sender, EventArgs args)
        { }

        protected abstract Binding Binding(long maxMessageSize, TimeSpan connectivityTimeout, TimeSpan debugTimeout);

        protected I innerChannel = default(I);
        public I Channel
        {
            get
            {
                lock (channelLock)
                {
                    if (!object.Equals(innerChannel, default(I)))
                    {
                        ICommunicationObject channelObject = innerChannel as ICommunicationObject;
                        if ((channelObject.State == CommunicationState.Faulted) || (channelObject.State == CommunicationState.Closed))
                        {
                            // Channel is faulted or closing for some reason, attempt to recreate channel
                            innerChannel = default(I);
                        }
                    }

                    if (object.Equals(innerChannel, default(I)))
                    {
                        Debug.Assert(Factory != null);
                        innerChannel = Factory.CreateChannel();
                        ((ICommunicationObject)innerChannel).Faulted += new EventHandler(Channel_Faulted);
                    }
                }

                return innerChannel;
            }
        }
        protected virtual void Channel_Faulted(object sender, EventArgs args)
        { }

        protected virtual void ValidateContract()
        { }

        protected void CloseChannel()
        {
            if (innerChannel == null)
            {
                return;
            }

            if (((IChannel)innerChannel).State == CommunicationState.Opened)
            {
                try
                {
                    ((IChannel)innerChannel).Close();
                }
                catch (TimeoutException /* timeout */)
                {
                    // Handle the timeout exception
                    ((IChannel)innerChannel).Abort();
                }
                catch (CommunicationException /* communicationException */)
                {
                    // Handle the communication exception
                    ((IChannel)innerChannel).Abort();
                }
            }
        }
        protected void CloseFactory()
        {
            if (Factory.State == CommunicationState.Opened)
            {
                try
                {
                    Factory.Close();
                }
                catch (TimeoutException /* timeout */)
                {
                    // Handle the timeout exception
                    Factory.Abort();
                }
                catch (CommunicationException /* communicationException */)
                {
                    // Handle the communication exception
                    Factory.Abort();
                }
            }
        }
        public virtual void Dispose()
        {
            CloseChannel();
            CloseFactory();
        }
    }
}