﻿using System;
using System.Collections.Generic;
using System.Collections.ObjectModel;
using System.Linq;
using System.ServiceModel;
using System.ServiceModel.Description;

namespace Company.Infrastructure.Policy
{
    [AttributeUsage(AttributeTargets.Class, Inherited = false, AllowMultiple = false)]
    public abstract class PolicyServiceBehaviorBase : Attribute, IServiceBehavior
    {
        /// <summary>
        /// Gets / Sets a type to be injected into the service's pipeline as a custom
        /// error handler. This type must implement 
        /// <see cref="System.ServiceModel.Dispatcher.IErrorHandler" />.
        /// </summary>
        public Type CustomErrorHandler
        { get; set; }

        protected Type EndpointBehavior
        { get; set; }

        protected IList<Type> OperationBehaviors
        { get; set; }

        public void AddBindingParameters(ServiceDescription serviceDescription, ServiceHostBase serviceHostBase, Collection<ServiceEndpoint> endpoints, System.ServiceModel.Channels.BindingParameterCollection bindingParameters)
        { }

        public void ApplyDispatchBehavior(ServiceDescription serviceDescription, ServiceHostBase serviceHostBase)
        {
            AddOperationAttributes(serviceDescription);
        }

        public void Validate(ServiceDescription serviceDescription, ServiceHostBase serviceHostBase)
        {
            EnforceServiceBehavior(serviceDescription, serviceHostBase);
        }

        protected virtual void AddOperationAttributes(ServiceDescription serviceDescription)
        {
            if (OperationBehaviors != null && OperationBehaviors.Any())
            {
                foreach (Type operationBehaviourType in OperationBehaviors)
                {
                    foreach (ServiceEndpoint endpoint in serviceDescription.Endpoints)
                    {
                        Func<OperationDescription, bool> findOperations = (operation) => !operation.Behaviors.Any(behavior => behavior.GetType().Equals(operationBehaviourType));
                        IEnumerable<OperationDescription> operations = endpoint.Contract.Operations.Where(findOperations);
                        foreach (OperationDescription operation in operations)
                        {
                            operation.Behaviors.Add((IOperationBehavior)Activator.CreateInstance(operationBehaviourType));
                        }
                    }
                }
            }
        }

        protected virtual void EnforceServiceBehavior(ServiceDescription serviceDescription, ServiceHostBase serviceHostBase)
        {
            ServiceBehaviorAttribute serviceBehavior = (ServiceBehaviorAttribute)serviceDescription.Behaviors.FirstOrDefault(behavior => behavior is ServiceBehaviorAttribute);
            if (serviceBehavior != null)
            {
                serviceBehavior.InstanceContextMode = InstanceContextMode.PerCall;
                serviceBehavior.ConcurrencyMode = ConcurrencyMode.Multiple;
                serviceBehavior.UseSynchronizationContext = false;
            }
        }
    }
}