﻿using System;
using System.Collections.Generic;
using System.ServiceModel.Description;
using System.ServiceModel.Security;
using Company.Infrastructure.ServiceModel.CustomConfiguration;
using System.Security.Principal;

namespace Company.Infrastructure.Security
{
    public class SecurityBehaviorProvider 
    {
        #region ServerBehaviors

        public static IEnumerable<IServiceBehavior> CreateServiceBehaviors(Type serviceType)
        {
            IList<IServiceBehavior> securityBehaviors = new List<IServiceBehavior>();
            SecurityType securityType = SecuritySettings.GetSecurityType();

            if (securityType != SecurityType.None)
            {
                ServiceAuthenticationBehavior authenticationBehavior = CreateServerAuthenticationBehavior();
                ServiceCredentials credentials = CreateServerCredentialBehavior(serviceType);
                ServiceAuthorizationBehavior authorizationBehavior = CreateServerAuthorizaionBehavior();

                securityBehaviors.Add(authenticationBehavior);
                securityBehaviors.Add(credentials);
                securityBehaviors.Add(authorizationBehavior);
            }

            return securityBehaviors;
        }

        private static ServiceAuthorizationBehavior CreateServerAuthorizaionBehavior()
        {
            ServiceAuthorizationBehavior authorizationBehavior = new ServiceAuthorizationBehavior();
            authorizationBehavior.PrincipalPermissionMode = PrincipalPermissionMode.UseWindowsGroups;
            authorizationBehavior.ImpersonateCallerForAllOperations = false;
            authorizationBehavior.ImpersonateOnSerializingReply = false;
            return authorizationBehavior;
        }

        private static ServiceCredentials CreateServerCredentialBehavior(Type serviceType)
        {
            string serverCertificateName = SecuritySettings.EnvironmentVariables.GetServiceHostServiceCertificateSubjectName(serviceType);
            string clientCertificateName = SecuritySettings.EnvironmentVariables.GetServiceHostClientCertificateSubjectName(serviceType);

            ServiceCredentials credentials = new ServiceCredentials();
            credentials.WindowsAuthentication.AllowAnonymousLogons = false;
            credentials.WindowsAuthentication.IncludeWindowsGroups = false;
            credentials.ServiceCertificate.Certificate = CertificateHelper.GetCertificate(serverCertificateName);
            credentials.ClientCertificate.Authentication.CertificateValidationMode = X509CertificateValidationMode.PeerOrChainTrust;
            credentials.ClientCertificate.Certificate = CertificateHelper.GetCertificate(clientCertificateName);
            return credentials;
        }

        private static ServiceAuthenticationBehavior CreateServerAuthenticationBehavior()
        {
            ServiceAuthenticationBehavior authenticationBehavior = new ServiceAuthenticationBehavior();
            authenticationBehavior.AuthenticationSchemes = System.Net.AuthenticationSchemes.IntegratedWindowsAuthentication;
            return authenticationBehavior;
        }

        #endregion

        #region EndPointBehaviors

        public static IEnumerable<IEndpointBehavior> CreateEndPointBehaviors(Type contract)
        {
            IList<IEndpointBehavior> securityBehaviors = new List<IEndpointBehavior>();
            SecurityType securityType = SecuritySettings.GetSecurityType();

            if (securityType != SecurityType.None)
            {
                ClientCredentials credentials = CreateEndPointCredentialBehavior(contract, securityType);

                securityBehaviors.Add(credentials);
            }

            return securityBehaviors;
        }

        private static ClientCredentials CreateEndPointCredentialBehavior(Type contract, SecurityType securityType)
        {
            string serverCertificateName = SecuritySettings.EnvironmentVariables.GetClientProxyServiceCertificateSubjectName(contract);
            string clientCertificateName = SecuritySettings.EnvironmentVariables.GetClientProxyClientCertificateSubjectName(contract);

            ClientCredentials credentials = new ClientCredentials();

            if (securityType == SecurityType.Impersonation)
            {
                credentials.Windows.AllowedImpersonationLevel = TokenImpersonationLevel.Impersonation;
            }
            else
            {
                credentials.Windows.AllowedImpersonationLevel = TokenImpersonationLevel.Delegation;
            }
            
            //credentials.Windows.AllowNtlm = true;
            credentials.ClientCertificate.Certificate = CertificateHelper.GetCertificate(clientCertificateName);
            credentials.ServiceCertificate.DefaultCertificate = CertificateHelper.GetCertificate(serverCertificateName);
            credentials.ServiceCertificate.Authentication.CertificateValidationMode = X509CertificateValidationMode.PeerOrChainTrust;
            return credentials;
        }

        #endregion
        
    }
}
