﻿using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Reflection;
using System.ServiceModel;
using System.ServiceModel.Configuration;
using Company.Infrastructure.Common;
using Company.Infrastructure.ServiceModel.CustomConfiguration;

namespace Company.Infrastructure.Server
{
    #region Event Args

    public abstract class HostArgs : EventArgs
    {
        public ServiceHost Host
        { get; internal set; }
    }
    public class OpeningArgs : HostArgs
    { }
    public class OpenedArgs : HostArgs
    { }
    public class ClosingArgs : HostArgs
    { }
    public class ClosedArgs : HostArgs
    { }
    public class FaultedArgs : EventArgs
    {
        public Exception Exception
        { get; internal set; }
    }

    #endregion

    public static class HostHelper
    {
        #region Constants

        const string NamspaceRoot = "Company";
        const string ServiceIdentifier = "Service";
        const string ComponentIdentifier = "Component";
        const string EndsWithIdentifies = "Manager";

        #endregion

        #region Event Handlers

        public static event EventHandler Initializing;
        public static event EventHandler Initialized;
        public static event EventHandler Launching;
        public static event EventHandler<OpeningArgs> Opening;
        public static event EventHandler<OpenedArgs> Opened;
        public static event EventHandler<ClosingArgs> Closing;
        public static event EventHandler<ClosedArgs> Closed;
        public static event EventHandler<FaultedArgs> Faulted;

        #endregion

        #region Public Members

        public static Dictionary<Type, object> Components = new Dictionary<Type, object>();

        #endregion

        #region Private Members

        static Assembly[] m_Assemblies = new Assembly[0];
        static List<ServiceHost> m_Hosts = new List<ServiceHost>();
        static List<Type> m_FailedToHost = new List<Type>();

        #endregion

        #region Properties

        public static IEnumerable<Type> Services { get; private set; }

        public static List<ServiceHost> Hosts
        {
            get
            {
                return m_Hosts;
            }
        }

        public static List<Type> FailedToHost
        {
            get
            {
                return m_FailedToHost;
            }
        }

        #endregion

        #region Public Methods

        public static void Start(Action launch)
        {
            if (launch == null)
            {
                throw new InvalidOperationException("Missing launch delegate.");
            }

            try
            {
                Initialize();
                Add();
                Validate();
                Open();
                AppDomain.CurrentDomain.ProcessExit += delegate { Close(); };
                RaiseLaunching();
                launch();
            }
            catch (Exception exception)
            {
                Close();
                throw exception;
            }
        }

        public static void End()
        {
            Close();
        }

        internal static void TestStart()
        {
            Initialize();
        }

        #endregion

        #region Sequenced Host Methods

        static void Initialize()
        {
            RaiseInitializing();
            IEnumerable<string> serviceNames = Settings.Services.Select(se => se.Name);
            if (serviceNames.Any())
            {
                List<Type> services = new List<Type>();
                foreach (string name in serviceNames)
                {
                    services.Add(FindService(name, ServiceIdentifier));
                }
                Services = services;
            }
            else
            {
                Services = FindServices(ServiceIdentifier);
            }

            if (!Services.Any())
            {
                throw new InvalidOperationException("No valid service types found in " + AppDomain.CurrentDomain.BaseDirectory);
            }

            List<Type> components = new List<Type>();
            components.AddRange(FindServices(ComponentIdentifier));
            Components = components.ToDictionary(k => k, k => (object)null);

            // TODO: Initialize process related utilities: logging, cache, etc...

            RaiseInitialized();
        }

        static void Add()
        {
            foreach (Type serviceType in Services)
            {
                ServiceHost host = SelectHost(serviceType);
                if (host != null)
                {
                    m_Hosts.Add(host);
                }
                else
                {
                    // If the host could not be created then store the service type in a list
                    FailedToHost.Add(serviceType);
                }
            }
        }

        static void Validate()
        {
            // DESIGN NOTE:
            // No need to validate service assembly references here. The FindService() methods called during initialze already automatically perform that check.

            IEnumerable<ChannelEndpointElement> clientEndpoints = Settings.ClientEndpoints;
            foreach (string contract in clientEndpoints.Select(endpoint => endpoint.Contract).Where(contract => !string.IsNullOrEmpty(contract)))
            {
                if (FindType(contract) == null)
                {
                    throw new FileNotFoundException("Missing contract type " + contract + ".");
                }
            }
        }

        static void Open()
        {
            foreach (ServiceHost host in m_Hosts)
            {
                RaiseOpening(host);
                host.Open();
                RaiseOpened(host);
            }
        }

        static void Close()
        {
            foreach (ServiceHost host in m_Hosts)
            {
                RaiseClosing(host);
                if (host.State == CommunicationState.Opened)
                {
                    host.Close();
                }
                else
                {
                    host.Abort();
                }
                RaiseClosed(host);
            }

            m_Hosts.Clear();

            // TODO: Close logs, clear caches, etc...
        }

        #endregion

        #region Raise Event Methods

        static void RaiseInitializing()
        {
            if (Initializing != null)
            {
                Initializing(null, null);
            }
        }

        static void RaiseInitialized()
        {
            if (Initialized != null)
            {
                Initialized(null, null);
            }
        }

        static void RaiseOpening(ServiceHost host)
        {
            if (Opening != null)
            {
                Opening(null, new OpeningArgs { Host = host });
            }
        }

        static void RaiseOpened(ServiceHost host)
        {
            if (Opened != null)
            {
                Opened(null, new OpenedArgs { Host = host });
            }
        }

        static void RaiseLaunching()
        {
            if (Launching != null)
            {
                Launching(null, null);
            }
        }

        static void RaiseClosing(ServiceHost host)
        {
            if (Closing != null)
            {
                Closing(null, new ClosingArgs { Host = host });
            }
        }

        static void RaiseClosed(ServiceHost host)
        {
            if (Closed != null)
            {
                Closed(null, new ClosedArgs { Host = host });
            }
        }

        static void RaiseFaulted(Exception ex)
        {
            if (Faulted != null)
            {
                Faulted(null, new FaultedArgs { Exception = ex });
            }
        }

        #endregion

        #region Private Methods

        private static ServiceHost SelectHost(Type serviceType)
        {
            PolicyServiceBehaviorBase policyBehaviorAttribute = null;

            // Load the policy to use for the service type from the config file
            PolicyAssemblyType policyAssemblyInfo = GetServiceHostPolicy(serviceType);

            if (policyAssemblyInfo == null)
            {
                IEnumerable<object> policyBehaviors = serviceType.GetCustomAttributes(typeof(PolicyServiceBehaviorBase), false);

                if (policyBehaviors.Count() > 1)
                {
                    throw new InvalidOperationException(string.Format("More than one policy has been defined for [{0}]", serviceType.ToString()));
                }

                policyBehaviorAttribute = (PolicyServiceBehaviorBase)policyBehaviors.FirstOrDefault();
            }
            else
            {
                Type serverPolicyAttributeType = null;
                try
                {
                    // See if the policy type are looking for exists in the appdomain allready
                    serverPolicyAttributeType = (from assembly in AppDomain.CurrentDomain.GetAssemblies()
                                                 from atype in assembly.GetTypes()
                                                 where atype.Name == policyAssemblyInfo.Type
                                                 select atype).FirstOrDefault();
                }
                catch (Exception)
                {
                }

                // The types does not exist in the app domain. Load the specified assembly into the app domain
                if (serverPolicyAttributeType == null)
                {
                    Assembly policyAssembly = Assembly.Load(policyAssemblyInfo.Assembly);
                    serverPolicyAttributeType = policyAssembly.GetType(policyAssemblyInfo.Type);
                }

                // The type exists in the appdomain, instantiate it.
                if (serverPolicyAttributeType != null)
                {
                    // Request the policy behaviour to create the host.
                    policyBehaviorAttribute = (PolicyServiceBehaviorBase)Activator.CreateInstance(serverPolicyAttributeType);
                }
            }

            ServiceHost serviceHost = null;

            // Create the host if there is a policy behavior attribute
            if (policyBehaviorAttribute != null)
            {
                // Load the base address from the config for the service
                string bindingName = policyBehaviorAttribute.GetBindingName();
                string baseAddress = Settings.EnvironmentVariables.GetServiceHostAddress(serviceType, bindingName, Settings.ServerName);
                var args = new Uri[] { new Uri(baseAddress) };

                serviceHost = policyBehaviorAttribute.CreateHost(serviceType, args);
            }

            return serviceHost;
        }

        private static IEnumerable<Type> FindServices(string identifier)
        {
            List<Type> serviceTypes = new List<Type>();
            IEnumerable<Assembly> assemblies = m_Assemblies.Where(assembly => assembly.FullName.StartsWith(NamspaceRoot + "." + identifier));
            foreach (Assembly assembly in assemblies)
            {
                serviceTypes.AddRange(assembly.GetExportedTypes().Where(type => type.FullName.EndsWith(EndsWithIdentifies)));
                //serviceTypes.AddRange(assembly.GetExportedTypes());
            }
            return serviceTypes.ToArray();
        }

        private static PolicyAssemblyType GetServiceHostPolicy(Type serviceType)
        {
            PolicyAssemblyType policy = new PolicyAssemblyType();

            ServiceHostElement serviceHostEntry = Settings.EnvironmentVariables.ServiceHosts.GetHostEntry(serviceType.ToString());
            string[] PolicyAssemblyArray = null;

            if (serviceHostEntry != null)
            {
                OverrideKeyValueElement policyOverrideElement = serviceHostEntry.GetEntry("ServerPolicy");

                if (policyOverrideElement != null)
                {
                    PolicyAssemblyArray = serviceHostEntry.GetEntry("ServerPolicy").Value.Split(',');
                    policy.Assembly = PolicyAssemblyArray[0];
                    policy.Type = PolicyAssemblyArray[1];
                }
            }
            else
            {
                OverrideKeyValueElement environmentEntry = Settings.EnvironmentVariables.Overrides.GetEntry("ServerPolicy");

                if (environmentEntry != null)
                {
                    PolicyAssemblyArray = environmentEntry.Value.Split(',');
                    policy.Assembly = PolicyAssemblyArray[0];
                    policy.Type = PolicyAssemblyArray[1];
                }
            }

            if (string.IsNullOrWhiteSpace(policy.Type) || string.IsNullOrWhiteSpace(policy.Assembly))
            {
                return null;
            }

            return policy;
        }

        private static void LoadAssemblies()
        {
            DirectoryInfo directories = new DirectoryInfo(AppDomain.CurrentDomain.BaseDirectory);
            FileInfo[] files = directories.GetFiles("*.dll").Where(f => f.Name.StartsWith(NamspaceRoot)).ToArray();

            foreach (FileInfo info in files)
            {
                try
                {
                    Assembly.LoadFile(info.FullName);
                }
                catch
                { }
            }

            m_Assemblies = AppDomain.CurrentDomain.GetAssemblies();
        }

        static HostHelper()
        {
            LoadAssemblies();
            Services = new Type[0] { };
        }

        static Type FindType(string fullName)
        {
            Type applicationType = null;
            foreach (var assembly in m_Assemblies.Where(assembly => assembly.FullName.StartsWith(NamspaceRoot)))
            {
                applicationType = assembly.ExportedTypes.FirstOrDefault(type => type.FullName.Equals(fullName));
                if (applicationType != null)
                {
                    break;
                }
            }
            return applicationType;
        }

        static Type FindService(string name, string identifier)
        {
            Type serviceType = null;
            foreach (var a in m_Assemblies.Where(a => a.FullName.StartsWith(NamspaceRoot + "." + identifier)))
            {
                serviceType = a.GetExportedTypes().Where(t => t.FullName.EndsWith(EndsWithIdentifies)).FirstOrDefault(t => t.FullName == name);
                //serviceType = a.GetExportedTypes().FirstOrDefault(t => t.FullName == name);
                if (serviceType != null)
                {
                    break;
                }
            }

            return serviceType;
        }

        #endregion
    }
}