﻿using System;
using System.Collections.Generic;
using System.Collections.ObjectModel;
using System.Linq;
using System.ServiceModel;
using System.ServiceModel.Description;

namespace Company.Infrastructure.Server
{
    [AttributeUsage(AttributeTargets.Class, Inherited = false, AllowMultiple = false)]
    public abstract class PolicyServiceBehaviorBase : Attribute, IServiceBehavior
    {
        #region Constructor

        public PolicyServiceBehaviorBase()
        {
        }

        #endregion

        #region Properties

        /// <summary>
        /// Gets / Sets a type to be injected into the service's pipeline as a custom
        /// error handler. This type must implement 
        /// <see cref="System.ServiceModel.Dispatcher.IErrorHandler" />.
        /// </summary>
        public Type CustomErrorHandler { get; set; }

        protected Type EndpointBehavior { get; set; }

        private IList<Type> _OperationBehaviors = new List<Type>();

        protected IList<Type> OperationBehaviors
        {
            get
            {
                return _OperationBehaviors;
            }
        }

        #endregion

        #region Public Methods

        public void AddBindingParameters(ServiceDescription serviceDescription, ServiceHostBase serviceHostBase, Collection<ServiceEndpoint> endpoints, System.ServiceModel.Channels.BindingParameterCollection bindingParameters)
        { }

        public void ApplyDispatchBehavior(ServiceDescription serviceDescription, ServiceHostBase serviceHostBase)
        {
            AddOperationAttributes(serviceDescription);
        }

        public void Validate(ServiceDescription serviceDescription, ServiceHostBase serviceHostBase)
        {
            EnforceServiceBehavior(serviceDescription, serviceHostBase);
        }

        #endregion

        #region Virtual Methods

        protected virtual void AddOperationAttributes(ServiceDescription serviceDescription)
        {
            if (OperationBehaviors != null && OperationBehaviors.Any())
            {
                foreach (Type operationBehaviourType in OperationBehaviors)
                {
                    foreach (ServiceEndpoint endpoint in serviceDescription.Endpoints)
                    {
                        Func<OperationDescription, bool> findOperations = (operation) => !operation.Behaviors.Any(behavior => behavior.GetType().Equals(operationBehaviourType));
                        IEnumerable<OperationDescription> operations = endpoint.Contract.Operations.Where(findOperations);
                        foreach (OperationDescription operation in operations)
                        {
                            operation.Behaviors.Add((IOperationBehavior)Activator.CreateInstance(operationBehaviourType));
                        }
                    }
                }
            }
        }

        #endregion

        #region Abstract Methods

        public abstract string GetBindingName();

        public abstract ServiceHost CreateHost(Type serviceType, System.Uri[] baseAddresses);

        protected abstract void EnforceServiceBehavior(ServiceDescription serviceDescription, ServiceHostBase serviceHostBase);

        #endregion
    }
}