﻿using Company.Infrastructure.Common;
using ServiceModelEx;
using System;
using System.Collections.Generic;
using System.Linq;
using System.ServiceModel;
using System.ServiceModel.Channels;
using System.ServiceModel.Configuration;
using System.ServiceModel.Description;

namespace Company.Infrastructure.Server
{
    public abstract class PolicyServiceHostBase<T> : ServiceHost<T>
    {
        #region Private Members

        bool m_OverrideAvailable = false;

        #endregion

        #region Constructors

        public PolicyServiceHostBase(params Uri[] baseAddresses)
            : base(baseAddresses)
        {
            Configure(Description);
        }

        public PolicyServiceHostBase(params string[] baseAddresses) : base(baseAddresses)
        {
            Configure(Description);
        }

        #endregion

        #region Properties

        protected string ServiceName
        {
            get
            {
                return Description.Name.Substring(Description.Name.LastIndexOf(".") + 1);
            }
        }

        protected virtual TimeSpan DefaultConnectivityTimeout
        {
            get
            {
                return BindingHelper.DefaultConnectivityTimeout;
            }
        }

        protected virtual TimeSpan DefaultDebugTimeout
        {
            get
            {
                return BindingHelper.DefaultDebugTimeout;
            }
        }

        protected virtual long DefaultMaxMessageSize
        {
            get
            {
                return BindingHelper.DefaultMaxMessageSize;
            }
        }

        #endregion

        #region Virtual Methods

        protected virtual void Host_Opened(object sender, EventArgs e) { }

        protected virtual void Host_Closed(object sender, EventArgs e) { }

        #endregion

        #region Override Methods

        // DESIGN NOTE:
        // - Control service-side configuration policy by overriding ApplyConfiguration()
        // - Must explicitly load configuration if overriding based on your configuration policy
        protected override void ApplyConfiguration()
        {
            m_OverrideAvailable = IsOverrideAvailable(Description.ServiceType);
            if (m_OverrideAvailable)
            {
                base.ApplyConfiguration();
            }
        }

        #endregion

        #region Protected Methods

        protected Type GetContract(string name, bool validate = true)
        {
            Type contract = Description.ServiceType.GetInterfaces().SingleOrDefault(i => i.Name == name);
            if ((validate) && (contract == null))
            {
                throw new InvalidOperationException("The service '" + ServiceName + "' does not implement the required interface '" + name + "'.");
            }

            return contract;
        }

        protected IEnumerable<Type> GetContracts()
        {
            return Description.ServiceType.GetInterfaces().Where(contract => contract.CustomAttributes.Any(attribute => attribute.AttributeType == typeof(ServiceContractAttribute)));
        }

        protected virtual void AddApplicationEndpoints(ServiceDescription description)
        {
            IEnumerable<ServiceEndpoint> overriddenEndpoints = Description.Endpoints.ToArray();
            foreach (Type contractType in GetContracts())
            {
                Binding binding = Binding(contractType);
                if (!overriddenEndpoints.Any(e => e.Binding.Scheme.Equals(binding.Scheme)))
                {
                    AddServiceEndpoint(contractType, binding, AddressHelper.EnforceEndpointAddress(contractType));
                }
            }
        }

        protected void EnforcePolicy(ServiceDescription description)
        {
            EnforceBehaviorPolicy(description);
            EnforceSecurityBehaviorPolicy(description);

            foreach (ServiceEndpoint endpoint in description.Endpoints)
            {
                // Enforce additional non-negotiable standard policy here
                EnforceEndPointPolicy(endpoint);
                EnforceEndPointSecurityPolicy(endpoint);

                if (m_OverrideAvailable)
                {
                    EnforcePolicyOnOverride(endpoint);
                }
            }
        }

        #endregion

        #region Private Methods

        void Configure(ServiceDescription description)
        {
            Opened += Host_Opened;
            Closed += Host_Closed;

            ApplyServiceConfiguration(description);
            EnforcePolicy(description);
        }

        void ApplyServiceConfiguration(ServiceDescription description)
        {
            AddStandardEndpoints(description);
            AddApplicationEndpoints(description);
        }

        void AddStandardEndpoints(ServiceDescription description) { }

        bool IsOverrideAvailable(Type serviceType)
        {
            IEnumerable<ServiceElement> servicesFromConfig = Settings.Services;
            return servicesFromConfig.Any(serviceElement => serviceElement.Name == serviceType.FullName);
        }

        void EnforcePolicyOnOverride(ServiceEndpoint endpoint)
        {
            // Enforce additional policy on override here.
            endpoint.Binding = BindingHelper.EnforceBindingPolicyOnFileOverride(endpoint.Contract.Name, endpoint.Binding);

            Uri baseAddress = BaseAddresses.First(uri => uri.Scheme.Equals(endpoint.Binding.Scheme));
            endpoint.Address = AddressHelper.EnforceApplicationEndpointAddress(baseAddress, endpoint.Contract.ContractType);
        }

        #endregion

        #region Abstract Methods

        protected abstract Binding Binding(Type contractType);

        protected abstract void EnforceBehaviorPolicy(ServiceDescription description);

        protected abstract void EnforceEndPointPolicy(ServiceEndpoint endpoint);

        protected abstract void EnforceEndPointSecurityPolicy(ServiceEndpoint endpoint);

        protected abstract void EnforceSecurityBehaviorPolicy(ServiceDescription description);

        #endregion
    }
}