﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.ServiceModel;
using System.Text;
using System.Threading.Tasks;

namespace Company.Infrastructure.ServiceModel.CustomConfiguration
{
    public static class CustomConfigurationExtensions
    {
        private const string _DefaultBinding = "net.tcp";
        private const string _DefaultHost = "localhost";
        private const string _DefaultPort = "808";

        public static OverrideKeyValueElement GetEntry(this OverrideKeyValueCollection collection, string key)
        {
            return collection.OfType<OverrideKeyValueElement>().Where(ov => ov.Key == key).FirstOrDefault();
        }

        public static ServiceHostElement GetHostEntry(this ServiceHostCollection collection, string type)
        {
            return collection.OfType<ServiceHostElement>().Where(h => h.Type == type).FirstOrDefault();
        }

        public static Service GetServiceEntry(this ServiceCollection collection, string contract)
        {
            return collection.OfType<Service>().Where(h => h.Contract == contract).FirstOrDefault();
        }

        public static Client GetClientEntry(this ClientCollection collection, string type)
        {
            return collection.OfType<Client>().Where(h => h.Type == type).FirstOrDefault();
        }

        public static string GetServiceHostServiceCertificateSubjectName(this EnvironmentConfig environmentConfig, Type serviceType)
        {
            ServiceHostElement serviceHostEntry = environmentConfig.ServiceHosts.GetHostEntry(serviceType.ToString());
            string certificateName;

            if (serviceHostEntry != null)
            {
                certificateName = serviceHostEntry.GetEntry("ServerCertificateSubjectName").Value;
            }
            else
            {
                certificateName = environmentConfig.Overrides.GetEntry("ServerCertificateSubjectName").Value;
            }

            return certificateName;
        }

        public static string GetServiceHostClientCertificateSubjectName(this EnvironmentConfig environmentConfig, Type serviceType)
        {
            ServiceHostElement serviceHostEntry = environmentConfig.ServiceHosts.GetHostEntry(serviceType.ToString());
            string certificateName;

            if (serviceHostEntry != null)
            {
                certificateName = serviceHostEntry.GetEntry("ClientCertificateSubjectName").Value;
            }
            else
            {
                certificateName = environmentConfig.Overrides.GetEntry("ClientCertificateSubjectName").Value;
            }

            return certificateName;
        }

        public static string GetClientProxyServiceCertificateSubjectName(this EnvironmentConfig environmentConfig, Type contract)
        {

            Company.Infrastructure.ServiceModel.CustomConfiguration.Client clientEntry = environmentConfig.Clients.GetClientEntry(contract.ToString());
            string certificateName;

            if (clientEntry != null)
            {
                certificateName = clientEntry.GetEntry("ServerCertificateSubjectName").Value;
            }
            else
            {
                certificateName = environmentConfig.Overrides.GetEntry("ServerCertificateSubjectName").Value;
            }

            return certificateName;
        }

        public static string GetClientProxyClientCertificateSubjectName(this EnvironmentConfig environmentConfig, Type contract)
        {
            Company.Infrastructure.ServiceModel.CustomConfiguration.Client clientEntry = environmentConfig.Clients.GetClientEntry(contract.ToString());
            string certificateName;

            if (clientEntry != null)
            {
                certificateName = clientEntry.GetEntry("ClientCertificateSubjectName").Value;
            }
            else
            {
                certificateName = environmentConfig.Overrides.GetEntry("ClientCertificateSubjectName").Value;
            }

            return certificateName;
        }

        public static string BindingToProtocolScheme(string binding)
        {
            switch (binding.ToLower())
            {
                case "netmsmqbinding":
                    return "net.msmq://";
                case "nettcpbinding":
                default:
                    return "net.tcp://";
            }
        }

        public static string GetServiceHostAddress(this EnvironmentConfig environmentConfig, Type serviceType, string bindingName, string serverName)
        {
            // Set the application defaults
            string binding = bindingName;
            string host = serverName;
            string port = string.Empty;

            // Check if there are "environment" overrides
            ServiceHostElement serviceHostEntry = environmentConfig.ServiceHosts.GetHostEntry(serviceType.ToString());
            if (serviceHostEntry != null)
            {
                binding = !string.IsNullOrWhiteSpace(serviceHostEntry.Binding) ? serviceHostEntry.Binding : binding;
                host = !string.IsNullOrWhiteSpace(serviceHostEntry.Host) ? serviceHostEntry.Host : host;
                port = !string.IsNullOrWhiteSpace(serviceHostEntry.Port) ? serviceHostEntry.Port : port;
            }

            // Check if there are "host" overrides
            OverrideKeyValueElement bindingOverride = environmentConfig.Overrides.GetEntry("Binding");
            OverrideKeyValueElement hostOverride = environmentConfig.Overrides.GetEntry("Host");
            OverrideKeyValueElement portOverride = environmentConfig.Overrides.GetEntry("Port");

            binding = bindingOverride != null ? bindingOverride.Value : bindingName;
            host = hostOverride != null ? hostOverride.Value : serverName;
            port = portOverride != null ? portOverride.Value : string.Empty;

            // Build the base address
            string baseAddress = string.Format("{0}{1}:{2}", BindingToProtocolScheme(binding), host, port);
            
            return baseAddress;
        }

        public static EndpointAddress GetClientEndpointAddress(this EnvironmentConfig environmentConfig, Type contract)
        {
            Client clientEntry = environmentConfig.Clients.GetClientEntry(contract.ToString());

            if (clientEntry != null)
            {
                return new EndpointAddress(clientEntry.Address);
            }

            return null;
        }

        public static Uri GetClientBaseAddress(this EnvironmentConfig environmentConfig, Type contract, string bindingName, string serverName)
        {
            Client clientEntry = environmentConfig.Clients.GetClientEntry(contract.ToString());

            // Set the application defaults
            string binding = bindingName;
            string host = serverName;
            string port = string.Empty;

            if (clientEntry != null)
            {
                binding = !string.IsNullOrWhiteSpace(clientEntry.Binding) ? clientEntry.Binding : binding;
                host = !string.IsNullOrWhiteSpace(clientEntry.Host) ? clientEntry.Host : host;
                port = !string.IsNullOrWhiteSpace(clientEntry.Port) ? clientEntry.Port : port;
            }

            // Check if there are "host" overrides
            OverrideKeyValueElement bindingOverride = environmentConfig.Overrides.GetEntry("Binding");
            OverrideKeyValueElement hostOverride = environmentConfig.Overrides.GetEntry("Host");
            OverrideKeyValueElement portOverride = environmentConfig.Overrides.GetEntry("Port");

            binding = bindingOverride != null ? bindingOverride.Value : bindingName;
            host = hostOverride != null ? hostOverride.Value : serverName;
            port = portOverride != null ? portOverride.Value : string.Empty;

            // Build the base address
            string baseAddress = string.Format("{0}{1}:{2}", BindingToProtocolScheme(binding), host, port);

            return new Uri(baseAddress);
        }
    }
}