﻿using System;
using System.Collections.Generic;
using System.Configuration;
using System.Linq;
using System.ServiceModel;
using System.ServiceModel.Channels;
using System.ServiceModel.Configuration;
using System.ServiceModel.Description;
using System.ServiceModel.Discovery;
using System.ServiceModel.Discovery.Configuration;
using ServiceModelEx;
using ServiceModelEx.TcpDiscovery;

namespace Assima.ServiceModel.Discovery
{
    public static class DiscoveryEndPointManager
    {
        /// <summary>
        /// Dictionary containing the types and addresses of previously discovered endpoints.
        /// </summary>
        private static Dictionary<Type, ServiceEndpoint> cachedEndPoints = new Dictionary<Type, ServiceEndpoint>();

        private static System.Timers.Timer cacheResetTimer = new System.Timers.Timer();


        //public static DiscoveryEndPointManager()
        //{

        //}

        /// <summary>
        /// Retrieves the endpoint for the specified channel from the config file. If none are found an attempt is made to get the endpoint from the discovery service.
        /// </summary>
        /// <typeparam name="TChannel">Contract / Interface type representing the channel signarure</typeparam>
        /// <param name="scope">Scope within which to search. This is only used by the discovery service</param>
        /// <returns>Returns the endpoint to the service</returns>
        public static ServiceEndpoint GetServiceAddress<TChannel>(Uri scope)
        {
            ClientSection clientSection = ConfigurationManager.GetSection("system.serviceModel/client") as ClientSection;

            if (clientSection != null)
            {
                foreach (ChannelEndpointElement endpointElement in clientSection.Endpoints)
                {
                    if (endpointElement.Contract == typeof(TChannel).ToString())
                    {
                        Binding binding = null;
                        string bindingConfigName = endpointElement.BindingConfiguration;

                        if (endpointElement.Address.Scheme == "net.tcp")
                        {
                            if (bindingConfigName.Length > 0)
                            {
                                binding = new NetTcpBinding(bindingConfigName);
                            }
                            else
                            {
                                binding = new NetTcpBinding();
                            }                            
                        }
                        if (endpointElement.Address.Scheme == "http")
                        {
                            if (bindingConfigName.Length > 0)
                            {
                                binding = new BasicHttpBinding(bindingConfigName);
                            }
                            else
                            {
                                binding = new BasicHttpBinding();
                            }                            
                        }
                        if (endpointElement.Address.Scheme == "net.msmq")
                        {
                            if (bindingConfigName.Length > 0)
                            {
                                binding = new NetMsmqBinding(bindingConfigName);
                            }
                            else
                            {
                                binding = new NetMsmqBinding();
                            }
                        }
                        if (binding != null)
                        {
                            ServiceEndpoint serviceEndpoint = new ServiceEndpoint(new ContractDescription(typeof(TChannel).ToString()), binding, new EndpointAddress(endpointElement.Address));
                            return serviceEndpoint;
                        }
                    }
                }
            }

            return DiscoveryAddress<TChannel>(scope, true);
        }

        /// <summary>
        /// Retrieves the Endpoint from the discovery service.
        /// </summary>
        /// <typeparam name="TChannel">Contract / Interface type representing the channel signarure</typeparam>
        /// <param name="scope">Scope within which to search. This is only used by the discovery service</param>        
        /// <param name="tcpBindingAsPreference">If true it will always return a tcp endpoint over any other endpoint type.</param>
        /// <returns>Returns the endpoint to the service</returns>
        public static ServiceEndpoint DiscoveryAddress<TChannel>(Uri scope, bool tcpBindingAsPreference = true)
        {
            lock (cachedEndPoints)
            {
                // Check the cache for previous discovered addresses.
                if (cachedEndPoints.ContainsKey(typeof(TChannel)))
                {
                    //Need to add code here to test if cached endpoint is still valid.

                    return cachedEndPoints[typeof(TChannel)];
                }

                ServiceEndpoint[] discoveredEndpoints = DiscoveryEndPointManager.GetServiceAddresses<TChannel>(scope);
                ServiceEndpoint discoveredEndpoint = discoveredEndpoints[0];

                // If more than one endpoint is returned and TcpBindingAsPreference is enabled then prioritize the return result for tcp endpoints - else what ever is first.
                if (!discoveredEndpoints.Any())
                {
                    throw new EndpointNotFoundException(string.Format("No endpoint for '{0}' could be discovered in scope '{1}'.", typeof(TChannel).Name, scope.ToString()));
                }
                else
                {
                    if (tcpBindingAsPreference)
                    {
                        discoveredEndpoint = discoveredEndpoints.FirstOrDefault(ep => ep.Binding.GetType() == typeof(NetTcpBinding));
                    }
                }

                cachedEndPoints.Add(typeof(TChannel), discoveredEndpoint);
                return discoveredEndpoint;
            }
        }

        /// <summary>
        /// Retrieves all the endpoints associated with the service representing the contract
        /// </summary>
        /// <typeparam name="TChannel">Contract / Interface type representing the channel signarure</typeparam>
        /// <param name="scope">Scope within which to search. This is only used by the discovery service</param>        
        /// <returns>Returns a list of discovered server endpoints</returns>
        public static ServiceEndpoint[] GetServiceAddresses<TChannel>(Uri scope = null)
        {
            DuplexDiscoveryClient discoveryClient = new DuplexDiscoveryClient();

            FindCriteria criteria = FindCriteria.CreateMetadataExchangeEndpointCriteria();
            criteria.MaxResults = 1;
            if (scope != null)
            {
                criteria.Scopes.Add(scope);
            }
            FindResponse discovered = discoveryClient.Find(criteria);

            if (discovered.Endpoints.Any())
            {
                Uri mexAddress = discovered.Endpoints[0].Address.Uri;
                ServiceEndpoint[] endpoints = MetadataHelper.GetEndpoints(mexAddress.AbsoluteUri, typeof(TChannel));

                if (endpoints.Any())
                {
                    return endpoints;
                }
            }

            throw new EndpointNotFoundException(string.Format("No endpoint for '{0}' could be discovered in scope '{1}'.", typeof(TChannel).Name, scope.ToString()));
        }

        /// <summary>
        /// Returns either the scopes defined in the configuration file or if there are none it creates a scope to the current machine.
        /// </summary>
        /// <returns>Returns a uri representing a unique scope for the host</returns>
        public static Uri GetHostServiceScope()
        {
            Uri scope = new Uri(string.Format("net.tcp://{0}/", Environment.MachineName));

            BehaviorsSection behaviours = ConfigurationManager.GetSection("system.serviceModel/behaviors") as BehaviorsSection;
            EndpointBehaviorElementCollection endPointBehaviours = behaviours.EndpointBehaviors;

            foreach (EndpointBehaviorElement endPointBehavourElement in endPointBehaviours)
            {
                if (endPointBehavourElement.ContainsKey("endpointDiscovery"))
                {
                    foreach (BehaviorExtensionElement behaviorExtension in endPointBehavourElement)
                    {
                        if (behaviorExtension.ConfigurationElementName == "endpointDiscovery")
                        {
                            EndpointDiscoveryElement discoveryBehaviorExtension = behaviorExtension as EndpointDiscoveryElement;
                            if (discoveryBehaviorExtension.Scopes.Count > 0)
                            {
                                return (discoveryBehaviorExtension.Scopes[0] as ScopeElement).Scope;
                            }
                        }
                    }
                }
            }

            return scope;
        }

        /// <summary>
        /// Retrieves the unique scope to use with the specified contract type.
        /// </summary>
        /// <param name="contractType">Contract type</param>
        /// <returns>Returns a uri representing a unique scope</returns>
        public static Uri GetScopeForService(Type contractType)
        {
            return new Uri(string.Format("{0}\\{1}", DiscoveryEndPointManager.GetHostServiceScope(), contractType.Name));
        }
    }
}
