﻿using ServiceModelEx.TcpDiscovery;
using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Linq;
using System.Reflection;
using System.ServiceModel;
using System.ServiceModel.Description;
using System.ServiceModel.Discovery;
using System.Text;
using System.Threading.Tasks;

namespace Company.Infrastructure.ServiceModel.Hosts
{
    public class ServiceHostManager
    {
        /// <summary>
        /// List of Services hosts that have been created within this process. 
        /// </summary>
        public List<ServiceHost> Hosts { get; set; }

        private string hostName;

        public ServiceHostManager(string HostName)
        {
            hostName = HostName;
            Hosts = new List<ServiceHost>();
        }

        public void OpenHosts()
        {
            List<ServiceHost> hostsToRemove = new List<ServiceHost>();
            bool failedToHost = false;

            // Iterate through the hosts and attempt to open them
            foreach (ServiceHost host in Hosts)
            {
                failedToHost = true;

                if (host != null)
                {
                    try
                    {
                        WriteToEventLog(EventLogEntryType.Warning, string.Format("Attempting to host service : '{0}'", host.Description.Name));
                        host.Open();
                        WriteToEventLog(EventLogEntryType.SuccessAudit, string.Format("Service '{0}' hosted.", host.Description.Name));

                        StringBuilder hostBehaviourDetails = ReadHostBehaviours(host);
                        Console.WriteLine(hostBehaviourDetails.ToString());
                        Trace.TraceInformation(hostBehaviourDetails.ToString());

                        if (Environment.UserInteractive)
                        {
                            Trace.TraceInformation(string.Format("Namespace : {0}", host.Description.Namespace));

                            foreach (ServiceEndpoint endPoint in host.Description.Endpoints)
                            {
                                string address = endPoint.Address.ToString();

                                StringBuilder endpointDetails = new StringBuilder();

                                endpointDetails.AppendLine(endPoint.Name);
                                endpointDetails.AppendLine(string.Format("Address: {0}", address));
                                endpointDetails.AppendLine(string.Format("Binding: {0} ({1})", endPoint.Binding.Name, endPoint.Binding.Scheme));
                                endpointDetails.AppendLine(string.Format("Contract: {0}", endPoint.Contract.Name));

                                ReadBindingSecurityConfiguration(endPoint, endpointDetails);

                                ReadDicoveryEndPointConfiguration(endPoint, endpointDetails);

                                Console.WriteLine(endpointDetails.ToString());
                                Trace.TraceInformation(endpointDetails.ToString());
                            }

                        }
                        failedToHost = false;
                    }
                    catch (ObjectDisposedException e)
                    {
                        WriteToEventLog(EventLogEntryType.Error, string.Format("Failed to host service : '{0}' - '{1}' ", host.Description.Name, e.ToString()));
                    }
                    catch (System.ServiceProcess.TimeoutException e)
                    {
                        WriteToEventLog(EventLogEntryType.Error, string.Format("Failed to host service : '{0}' - '{1}' ", host.Description.Name, e.ToString()));
                    }
                    catch (CommunicationObjectFaultedException e)
                    {
                        WriteToEventLog(EventLogEntryType.Error, string.Format("Failed to host service : '{0}' - '{1}' ", host.Description.Name, e.ToString()));
                    }
                    catch (InvalidOperationException e)
                    {
                        WriteToEventLog(EventLogEntryType.Error, string.Format("Failed to host service : '{0}' - '{1}' ", host.Description.Name, e.ToString()));
                    }
                }

                if (failedToHost)
                {
                    hostsToRemove.Add(host);
                }
                Console.WriteLine();
            }

            // Remove the hosts that did not open
            foreach (ServiceHost host in hostsToRemove)
            {
                Hosts.Remove(host);
                Debugger.Log(0, "WCF Hosts", string.Format("Unable to load : '{0}'", hostName));
            }

            if (hostsToRemove.Count == 0)
            {
                WriteToEventLog(EventLogEntryType.SuccessAudit, string.Format("'{0}' Started.", hostName));
            }
        }

        public static StringBuilder ReadHostBehaviours(ServiceHost host)
        {
            StringBuilder hostBehaviourDetails = new StringBuilder();

            if (host.Description.Behaviors.Contains(typeof(ServiceBehaviorAttribute)))
            {
                ServiceBehaviorAttribute behaviour = host.Description.Behaviors[typeof(ServiceBehaviorAttribute)] as ServiceBehaviorAttribute;
                hostBehaviourDetails.AppendLine(string.Format("Behaviour: ServiceBehaviorAttribute"));
                hostBehaviourDetails.AppendLine(string.Format("Instance Context Mode: {0}", behaviour.InstanceContextMode.ToString()));
                hostBehaviourDetails.AppendLine(string.Format("Concurrency Mode: {0}", behaviour.ConcurrencyMode.ToString()));
                hostBehaviourDetails.AppendLine(string.Format("Transaction Isolation Level: {0}", behaviour.TransactionIsolationLevel.ToString()));
                hostBehaviourDetails.AppendLine(string.Format("Include ExceptionDetail In Faults: {0}", behaviour.IncludeExceptionDetailInFaults.ToString()));
                if (behaviour.Namespace != null)
                {
                    hostBehaviourDetails.AppendLine(string.Format("Namespace: {0}", behaviour.Namespace.ToString()));
                }
                hostBehaviourDetails.AppendLine();
            }

            if (host.Description.Behaviors.Contains(typeof(ServiceAuthenticationBehavior)))
            {
                ServiceAuthenticationBehavior behaviour = host.Description.Behaviors[typeof(ServiceAuthenticationBehavior)] as ServiceAuthenticationBehavior;
                hostBehaviourDetails.AppendLine(string.Format("Behaviour: {0}", "ServiceAuthenticationBehavior"));
                hostBehaviourDetails.AppendLine(string.Format("Authentication Scheme: {0}", behaviour.AuthenticationSchemes.ToString()));
                hostBehaviourDetails.AppendLine();
            }

            if (host.Description.Behaviors.Contains(typeof(ServiceAuthorizationBehavior)))
            {
                ServiceAuthorizationBehavior behaviour = host.Description.Behaviors[typeof(ServiceAuthorizationBehavior)] as ServiceAuthorizationBehavior;
                hostBehaviourDetails.AppendLine(string.Format("Behaviour: {0}", "ServiceAuthorizationBehavior"));
                hostBehaviourDetails.AppendLine(string.Format("Impersonate Caller For All Operations: {0}", behaviour.ImpersonateCallerForAllOperations.ToString()));
                hostBehaviourDetails.AppendLine(string.Format("Impersonate On Serializing Reply: {0}", behaviour.ImpersonateOnSerializingReply.ToString()));
                hostBehaviourDetails.AppendLine(string.Format("Principal Permission Mode: {0}", behaviour.PrincipalPermissionMode.ToString()));
                hostBehaviourDetails.AppendLine();
            }

            if (host.Description.Behaviors.Contains(typeof(ServiceCredentials)))
            {
                ServiceCredentials behaviour = host.Description.Behaviors[typeof(ServiceCredentials)] as ServiceCredentials;
                hostBehaviourDetails.AppendLine(string.Format("Behaviour: {0}", "ServiceCredentials"));

                hostBehaviourDetails.AppendLine(string.Format("Windows Authentication:"));
                hostBehaviourDetails.AppendLine(string.Format("Allow Anonymous Logons: {0}", behaviour.WindowsAuthentication.AllowAnonymousLogons.ToString()));
                hostBehaviourDetails.AppendLine(string.Format("Include Windows Groups: {0}", behaviour.WindowsAuthentication.IncludeWindowsGroups.ToString()));
                hostBehaviourDetails.AppendLine(string.Format("Use Identity Configuration: {0}", behaviour.UseIdentityConfiguration.ToString()));
                hostBehaviourDetails.AppendLine(string.Format("UserName Password Validation Mode: {0}", behaviour.UserNameAuthentication.UserNamePasswordValidationMode.ToString()));
                hostBehaviourDetails.AppendLine();
                hostBehaviourDetails.AppendLine(string.Format("Authentication:"));
                hostBehaviourDetails.AppendLine(string.Format("Include Windows Groups: {0}", behaviour.ClientCertificate.Authentication.IncludeWindowsGroups.ToString()));
                hostBehaviourDetails.AppendLine(string.Format("Revocation Mode: {0}", behaviour.ClientCertificate.Authentication.RevocationMode.ToString()));
                hostBehaviourDetails.AppendLine(string.Format("Trusted Store Location: {0}", behaviour.ClientCertificate.Authentication.TrustedStoreLocation.ToString()));
                hostBehaviourDetails.AppendLine();

                if (behaviour.ClientCertificate.Certificate != null)
                {
                    hostBehaviourDetails.AppendLine(string.Format("Client Certificate:"));
                    hostBehaviourDetails.AppendLine(string.Format("FriendlyName: {0}", behaviour.ClientCertificate.Certificate.FriendlyName.ToString()));
                    hostBehaviourDetails.AppendLine(string.Format("SubjectName: {0}", behaviour.ClientCertificate.Certificate.SubjectName.Name));
                    hostBehaviourDetails.AppendLine(string.Format("SerialNumber: {0}", behaviour.ClientCertificate.Certificate.SerialNumber.ToString()));
                    hostBehaviourDetails.AppendLine(string.Format("Thumbprint: {0}", behaviour.ClientCertificate.Certificate.Thumbprint.ToString()));
                }
                if (behaviour.ServiceCertificate.Certificate != null)
                {
                    hostBehaviourDetails.AppendLine();
                    hostBehaviourDetails.AppendLine(string.Format("Server Certificate:"));
                    hostBehaviourDetails.AppendLine(string.Format("FriendlyName: {0}", behaviour.ServiceCertificate.Certificate.FriendlyName.ToString()));
                    hostBehaviourDetails.AppendLine(string.Format("SubjectName: {0}", behaviour.ServiceCertificate.Certificate.SubjectName.Name));
                    hostBehaviourDetails.AppendLine(string.Format("SerialNumber: {0}", behaviour.ServiceCertificate.Certificate.SerialNumber.ToString()));
                    hostBehaviourDetails.AppendLine(string.Format("Thumbprint: {0}", behaviour.ServiceCertificate.Certificate.Thumbprint.ToString()));
                    hostBehaviourDetails.AppendLine();
                }                
            }
            return hostBehaviourDetails;
        }

        private static void ReadDicoveryEndPointConfiguration(ServiceEndpoint endPoint, StringBuilder endpointDetails)
        {
            if (endPoint.EndpointBehaviors.Contains(typeof(EndpointDiscoveryBehavior)))
            {
                EndpointDiscoveryBehavior discoveryBehaviour = endPoint.EndpointBehaviors[typeof(EndpointDiscoveryBehavior)] as EndpointDiscoveryBehavior;

                if (discoveryBehaviour != null)
                {
                    foreach (Uri scope in discoveryBehaviour.Scopes)
                    {
                        endpointDetails.AppendLine(string.Format("Scope: {0}", scope.ToString()));
                    }
                }
            }
        }

        public static void ReadBindingSecurityConfiguration(ServiceEndpoint endPoint, StringBuilder endpointDetails)
        {
            if (endPoint.Binding is NetTcpBinding)
            {
                NetTcpBinding binding = endPoint.Binding as NetTcpBinding;

                endpointDetails.AppendLine(string.Format("Binding Security Mode: {0}", binding.Security.Mode.ToString()));
                if (binding.Security.Mode != SecurityMode.None)
                {
                    endpointDetails.AppendLine(string.Format("Transport Credential Type: {0}", binding.Security.Transport.ClientCredentialType.ToString()));
                    endpointDetails.AppendLine(string.Format("Transport Protection Level: {0}", binding.Security.Transport.ProtectionLevel.ToString()));
                    endpointDetails.AppendLine(string.Format("Message Credential Type: {0}", binding.Security.Message.ClientCredentialType.ToString()));
                }
            }

            if (endPoint.Binding is NetHttpBinding)
            {
                NetHttpBinding binding = endPoint.Binding as NetHttpBinding;

                endpointDetails.AppendLine(string.Format("Binding Security Mode: {0}", binding.Security.Mode.ToString()));
                if (binding.Security.Mode != BasicHttpSecurityMode.None)
                {
                    endpointDetails.AppendLine(string.Format("Transport Credential Type: {0}", binding.Security.Transport.ClientCredentialType.ToString()));
                    endpointDetails.AppendLine(string.Format("Message Credential Type: {0}", binding.Security.Message.ClientCredentialType.ToString()));
                }
            }

            if (endPoint.Binding is NetHttpsBinding)
            {
                NetHttpsBinding binding = endPoint.Binding as NetHttpsBinding;

                endpointDetails.AppendLine(string.Format("Binding Security Mode: {0}", binding.Security.Mode.ToString()));
                endpointDetails.AppendLine(string.Format("Transport Credential Type: {0}", binding.Security.Transport.ClientCredentialType.ToString()));
                endpointDetails.AppendLine(string.Format("Message Credential Type: {0}", binding.Security.Message.ClientCredentialType.ToString()));

            }

            if (endPoint.Binding is NetMsmqBinding)
            {
                NetMsmqBinding binding = endPoint.Binding as NetMsmqBinding;

                endpointDetails.AppendLine(string.Format("Binding Security Mode: {0}", binding.Security.Mode.ToString()));
                if (binding.Security.Mode != NetMsmqSecurityMode.None)
                {
                    endpointDetails.AppendLine(string.Format("Transport Credential Type: {0}", binding.Security.Transport.MsmqAuthenticationMode.ToString()));
                    endpointDetails.AppendLine(string.Format("Transport Protection Level: {0}", binding.Security.Transport.MsmqProtectionLevel.ToString()));
                    endpointDetails.AppendLine(string.Format("Message Credential Type: {0}", binding.Security.Message.ClientCredentialType.ToString()));
                }
            }
        }

        public void CloseHosts()
        {
            //WriteToEventLog(EventLogEntryType.Warning, string.Format("Closing host: '{0}'", hostName));

            foreach (ServiceHost host in Hosts)
            {
                if (host != null)
                {
                    try
                    {
                        host.Close();
                    }
                    catch (ObjectDisposedException e)
                    {
                        WriteToEventLog(EventLogEntryType.Error, string.Format("Failed to close service : '{0}' - '{1}' ", host.Description.Name, e.ToString()));
                        throw;
                    }
                    catch (System.ServiceProcess.TimeoutException e)
                    {
                        WriteToEventLog(EventLogEntryType.Error, string.Format("Failed to close service : '{0}' - '{1}' ", host.Description.Name, e.ToString()));
                        throw;
                    }
                    catch (CommunicationObjectFaultedException e)
                    {
                        WriteToEventLog(EventLogEntryType.Error, string.Format("Failed to close service : '{0}' - '{1}' ", host.Description.Name, e.ToString()));
                        throw;
                    }
                    catch (InvalidOperationException e)
                    {
                        WriteToEventLog(EventLogEntryType.Error, string.Format("Failed to close service : '{0}' - '{1}' ", host.Description.Name, e.ToString()));
                        throw;
                    }
                }
            }
        }

        /// <summary>
        /// Wrapper method use to write to the event log.
        /// </summary>
        /// <param name="logType"></param>
        /// <param name="logEntry"></param>
        private void WriteToEventLog(EventLogEntryType logType, string logEntry)
        {
            Trace.TraceInformation(logEntry);
            try
            {
                EventLog log = new EventLog();
                log.Source = hostName;
                log.WriteEntry(logEntry, logType);
            }
            catch (Exception)
            {
                // Swallow the exception.   
            }

            switch (logType)
            {
                case EventLogEntryType.Warning:
                    Console.ForegroundColor = ConsoleColor.Yellow;
                    break;

                case EventLogEntryType.Error:
                    Console.ForegroundColor = ConsoleColor.Red;
                    break;

                case EventLogEntryType.SuccessAudit:
                    Console.ForegroundColor = ConsoleColor.Green;
                    break;

                default:
                    Console.ForegroundColor = ConsoleColor.White;
                    break;
            }

            Console.WriteLine(string.Format("{0} - {1}", logType.ToString(), logEntry));
            Console.ForegroundColor = ConsoleColor.White;
        }

        //public static ServiceHost CreateHost<ServiceType, ServiceInterface>(bool useDiscovery)
        //{
        //    if (useDiscovery)
        //    {
        //        Uri scope = DiscoveryEndPointManager.GetScopeForService(typeof(ServiceInterface));

        //        MethodInfo createHostMethod  = typeof(DiscoveryFactory).GetMethod("CreateDiscoverableHost");
        //        MethodInfo genericMethod = createHostMethod.MakeGenericMethod(typeof(ServiceType));

        //        return genericMethod.Invoke(null, new object[1] { scope }) as ServiceHost;
        //    }
        //    else
        //    {
        //        return new ServiceHost(typeof(ServiceType));
        //    }
        //}

    }
}
