﻿using System;
using System.Collections;
using System.Collections.Generic;
using System.Linq;
using System.ServiceModel;
using System.Text;
using System.Threading.Tasks;
using Company.Infrastructure.ServiceModel.Errors;

namespace Company.Infrastructure.ServiceModel.Utils
{
    /// <summary>
    /// Helper class for asserting various conditions, raises a fault exception when the condition is not met
    /// </summary>
    public static class Assert
    {
		/// <summary>
		/// Checks that all items in the collection match the expression
		/// </summary>
		/// <typeparam name="T">The type of the item in the enumerable</typeparam>
		/// <param name="enumerable">The enumerable to check</param>
		/// <param name="expression">The expression to execute against each item in the enumerable</param>
		/// <param name="faultCode">Unique error code if the condition is not met.</param>
		/// <param name="faultReason">Description of the error if the condition is not met. If this is not supplied then a default message will be used.</param>
		/// <param name="variableName">The name of the variable being checked, only used to generate the default message if no fault reason is supplied</param>
		public static void All<T>(IEnumerable<T> enumerable, Func<T, bool> expression, int faultCode, string faultReason = "", string variableName = "")
		{
			if (enumerable == null || enumerable.Any(item => !expression(item)))
			{
				// Build the default message
				if (string.IsNullOrWhiteSpace(faultReason))
				{
					variableName = !string.IsNullOrWhiteSpace(variableName) ? variableName + " " : string.Empty;
					faultReason = string.Format("Not all items in the enumerable {0} match the supplied expression.", variableName);
				}

				RaiseFaultException(faultReason, faultCode);
			}
		}

		/// <summary>
		/// Verifies that two specified generic type data are equal by using the equality operator. The assertion fails if they are not equal.
		/// </summary>
		/// <typeparam name="T">The type of the items to check</typeparam>
		/// <param name="expected">The expected value</param>
		/// <param name="actual">The actual value</param>
		/// <param name="faultCode">Unique error code if the condition is not met.</param>
		/// <param name="faultReason">Description of the error if the condition is not met. If this is not supplied then a default message will be used.</param>		 
		public static void AreEqual<T>(T expected, T actual, int faultCode, string faultReason = "")
		{
			if (!object.Equals((object)expected, (object)actual))
			{
				// Build the default message
				if (string.IsNullOrWhiteSpace(faultReason))
					faultReason = string.Format("Actual result and expected result are not equal. Expected: <{0}>. Actual: <{1}>.", expected, actual);

				RaiseFaultException(faultReason, faultCode);
			}
		}

		/// <summary>
		/// Verifies that two specified generic type data are not equal. The assertion fails if they are equal. 
		/// </summary>
		/// <typeparam name="T">The type of the items to check</typeparam>   
		/// <param name="notExpected">The value the actual value should not be equal to</param>
		/// <param name="actual">The actual value</param>
		/// <param name="faultCode">Unique error code if the condition is not met.</param>
		/// <param name="faultReason">Description of the error if the condition is not met. If this is not supplied then a default message will be used.</param>
		public static void AreNotEqual<T>(T notExpected, T actual, int faultCode, string faultReason = "")
		{
			if (object.Equals((object)notExpected, (object)actual))
			{
				// Build the default message
				if (string.IsNullOrWhiteSpace(faultReason))
					faultReason = string.Format("Expected any value except:<{0}>. Actual:<{1}>.", notExpected, actual);

				RaiseFaultException(faultReason, faultCode);
			}
		}

		/// <summary>
		/// Checks if an enumerable contains a specific value
		/// </summary>
		/// <typeparam name="T">The type of the item in the enumerable</typeparam>
		/// <param name="enumerable">The enumerable to check</param>
		/// <param name="value">The value to for</param>
		/// <param name="faultCode">Unique error code if the condition is not met.</param>
		/// <param name="faultReason">Description of the error if the condition is not met. If this is not supplied then a default message will be used.</param>
		/// <param name="variableName">The name of the variable being checked, only used to generate the default message if no fault reason is supplied</param>
		public static void Contains<T>(IEnumerable<T> enumerable, T value, int faultCode,string faultReason = "", string variableName = "")
		{
			if (enumerable == null || !enumerable.Contains(value))
			{
				// Build the default message
				if (string.IsNullOrWhiteSpace(faultReason))
				{
					variableName = !string.IsNullOrWhiteSpace(variableName) ? variableName : "Object";
					faultReason = string.Format("{0} does not contain the value '{1}'.", variableName, value);
				}

				RaiseFaultException(faultReason, faultCode);
			}
		}

		/// <summary>
		/// Checks if a number is greater than a given value
		/// </summary>
		/// <param name="checkValue">The value to check</param>
		/// <param name="greaterThanValue">The value that the value to check must be greater than </param>
		/// <param name="faultCode">Unique error code if the condition is not met.</param>
		/// <param name="faultReason">Description of the error if the condition is not met. If this is not supplied then a default message will be used.</param>
		public static void GreaterThan(int checkValue, int greaterThanValue, int faultCode, string faultReason = "")
		{
			if (checkValue <= greaterThanValue)
			{
				// Build the default message
				if (string.IsNullOrWhiteSpace(faultReason))
					faultReason = string.Format("The value <{0}> is not greater than <{1}>.", checkValue, greaterThanValue);

				RaiseFaultException(faultReason, faultCode);
			}
		}

		/// <summary>
		/// Checks if a number is greater than a given value
		/// </summary>
		/// <param name="checkValue">The value to check</param>
		/// <param name="greaterThanValue">The value that the value to check must be greater than </param>
		/// <param name="faultCode">Unique error code if the condition is not met.</param>
		/// <param name="faultReason">Description of the error if the condition is not met. If this is not supplied then a default message will be used.</param>
		public static void GreaterThan(long checkValue, long greaterThanValue, int faultCode, string faultReason = "")
		{
			if (checkValue <= greaterThanValue)
			{
				// Build the default message
				if (string.IsNullOrWhiteSpace(faultReason))
					faultReason = string.Format("The value <{0}> is not greater than <{1}>.", checkValue, greaterThanValue);

				RaiseFaultException(faultReason, faultCode);
			}
		}

		/// <summary>
		/// Checks if a boolean is false
		/// </summary>
		/// <param name="value">The value to check</param>
		/// <param name="faultCode">Unique error code if the condition is not met.</param>
		/// <param name="faultReason">Description of the error if the condition is not met. If this is not supplied then a default message will be used.</param>
		/// <param name="variableName">The name of the variable being checked, only used to generate the default message if no fault reason is supplied</param>
		public static void IsFalse(bool value, int faultCode,string faultReason = "", string variableName = "")
		{
			if (value)
			{
				// Build the default message
				if (string.IsNullOrWhiteSpace(faultReason))
				{
					variableName = !string.IsNullOrWhiteSpace(variableName) ? variableName : "Object";
					faultReason = string.Format("{0} cannot be true.", variableName);
				}

				RaiseFaultException(faultReason, faultCode);
			}
		}

        /// <summary>
        /// Checks if a value is in the specified range
        /// </summary>
        /// <param name="checkValue">The value to check</param>
        /// <param name="minimum">Minimum value of the range</param>
        /// <param name="maximum">Macximum value of the range</param>
        /// <param name="inclusive">If true then the minimum and maximum values are included in the range of valid values.</param>
        /// <param name="faultCode">Unique error code if the condition is not met.</param>
        /// <param name="faultReason">Description of the error if the condition is not met. If this is not supplied then a default message will be used.</param>
        public static void IsInRange(int checkValue, int minimum, int maximum, bool inclusive, int faultCode, string faultReason = "")
        {
            int minValue = inclusive ? minimum : minimum + 1;
            int maxValue = inclusive ? maximum : maximum - 1;

            if (checkValue < minValue || checkValue > maxValue)
            {
                // Build the default message
                if (string.IsNullOrWhiteSpace(faultReason))
                    faultReason = string.Format("The value <{0}> is not in the range ({1}, {2}.", checkValue, minValue, maxValue);

                RaiseFaultException(faultReason, faultCode);
            }
        }

		/// <summary>
		/// Checks that an enumerable is not an empty list
		/// </summary>
		/// <typeparam name="T">The type of the item in the enumerable</typeparam>
		/// <param name="enumerable">The enumerable to check</param>
		/// <param name="faultCode">Unique error code if the condition is not met.</param>
		/// <param name="faultReason">Description of the error if the condition is not met. If this is not supplied then a default message will be used.</param>
		/// <param name="variableName">The name of the variable being checked, only used to generate the default message if no fault reason is supplied</param>
		public static void IsNotEmpty<T>(IEnumerable<T> enumerable, int faultCode,string faultReason = "", string variableName = "")
		{
			if (enumerable == null || !enumerable.Any())
			{
				// Build the default message
				if (string.IsNullOrWhiteSpace(faultReason))
				{
					variableName = !string.IsNullOrWhiteSpace(variableName) ? variableName : "Object";
					faultReason = string.Format("{0} cannot be empty.", variableName);
				}

				RaiseFaultException(faultReason, faultCode);
			}
		}

		/// <summary>
		/// Checks that an object is not null
		/// </summary>
		/// <param name="instance">The object instance to check</param>		
		/// <param name="faultCode">Unique error code if the condition is not met.</param>
		/// <param name="faultReason">Description of the error if the condition is not met. If this is not supplied then a default message will be used.</param>
		/// <param name="variableName">The name of the variable being checked, only used to generate the default message if no fault reason is supplied</param>
		public static void IsNotNull(object instance, int faultCode, string faultReason = "", string variableName = "")
		{
			if (instance == null)
			{
				// Build the default message
				if (string.IsNullOrWhiteSpace(faultReason))
				{
					variableName = !string.IsNullOrWhiteSpace(variableName) ? variableName : "Object";
					faultReason = string.Format("{0} cannot be null.", variableName);
				}

				RaiseFaultException(faultReason, faultCode);
			}
		}

		/// <summary>
		/// Checks that an object is null
		/// </summary>
		/// <param name="instance">The object instance to check</param>
		/// <param name="faultCode">Unique error code if the condition is not met.</param>
		/// <param name="faultReason">Description of the error if the condition is not met. If this is not supplied then a default message will be used.</param>
		/// <param name="variableName">The name of the variable being checked, only used to generate the default message if no fault reason is supplied</param>
		public static void IsNull(object instance, int faultCode,string faultReason = "", string variableName = "")
		{
			if (instance != null)
			{
				// Build the default message
				if (string.IsNullOrWhiteSpace(faultReason))
				{
					variableName = !string.IsNullOrWhiteSpace(variableName) ? variableName : "Object";
					faultReason = string.Format("{0} must be null.", variableName);
				}

				RaiseFaultException(faultReason, faultCode);
			}
		}

		/// <summary>
		/// Checks that a boolean value is true
		/// </summary>
		/// <param name="value">The value to check</param>
		/// <param name="faultCode">Unique error code if the condition is not met.</param>
		/// <param name="faultReason">Description of the error if the condition is not met. If this is not supplied then a default message will be used.</param>
		/// <param name="variableName">The name of the variable being checked, only used to generate the default message if no fault reason is supplied</param>
		public static void IsTrue(bool value, int faultCode, string faultReason = "", string variableName = "")
		{
			if (!value)
			{
				// Build the default message
				if (string.IsNullOrWhiteSpace(faultReason))
				{
					variableName = !string.IsNullOrWhiteSpace(variableName) ? variableName : "Object";
					faultReason = string.Format("{0} cannot be false.", variableName);
				}

				RaiseFaultException(faultReason, faultCode);
			}
		}

		/// <summary>
		/// Raises a fault exception with a given message
		/// </summary>
		/// <param name="faultCode">Unique error code if the condition is not met.</param>
		/// <param name="faultReason">Description of the error if the condition is not met. If this is not supplied then a default message will be used.</param>		 
		private static void RaiseFaultException(string faultReason, int faultCode)
		{
			// Get the stack trace info
			string methodName = GetCallerName();
			string className = GetCallerType();

			// Append the stack trace info
			if (!string.IsNullOrEmpty(methodName) && !string.IsNullOrEmpty(className))
				faultReason = string.Format("{0}{1} (Source: {2}.{3})", faultReason, !faultReason.EndsWith(".") ? "." : string.Empty, className, methodName);

			// Just in case
			if (faultReason == null)
				faultReason = string.Empty;

			// Raise the exception			
			FaultReason reason = new FaultReason(faultReason);			
			FaultCode code = new FaultCode(faultCode.ToString());

			ErrorDetail detail = new ErrorDetail();
			detail.Message = faultReason;
			detail.Code = faultCode;

			throw new FaultException<ErrorDetail>(detail, reason, code);
		}


		/// <summary>
		/// Gets the name of the method that called the assertion
		/// </summary>
		/// <returns>Returns the name of the method that called the assertion</returns>
		private static string GetCallerName()
		{
			System.Reflection.MethodBase method = GetCallingMethod();

			if (method != null)
				return method.Name;

			return string.Empty;
		}

		/// <summary>
		/// Gets the class name of the method that called the assertion
		/// </summary>
		/// <returns>Returns the class name of the method that called the assertion</returns>
		private static string GetCallerType()
		{
			System.Reflection.MethodBase method = GetCallingMethod();

			if (method != null)
				return method.DeclaringType.Name;

			return string.Empty;
		}

		/// <summary>
		/// Gets the method that called the assertion
		/// </summary>
		/// <returns>Returns the method that called the assertion</returns>
		private static System.Reflection.MethodBase GetCallingMethod()
		{
			System.Diagnostics.StackTrace stackTrace = new System.Diagnostics.StackTrace();
			System.Diagnostics.StackFrame[] frames = stackTrace.GetFrames();

			if (frames != null && frames.Any())
			{
				foreach (System.Diagnostics.StackFrame frame in frames)
				{
					System.Reflection.MethodBase method = frame.GetMethod();
					if (method != null && method.DeclaringType != typeof(Assert))
						return method;
				}
			}

			return null;
		}
    }
}
