//  2011 IDesign Inc. All rights reserved 
//Questions? Comments? go to 
//http:www.idesign.net

using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Linq;
using System.Runtime.CompilerServices;
using System.Runtime.Serialization;
using System.Security.Claims;
using System.Security.Principal;
using System.ServiceModel;
using System.Threading;
using System.Web;
using ServiceModelEx.Security.Principal;

namespace ServiceModelEx
{
    [DataContract]
    public class SecurityCallStack
    {
        #region Fields

        [DataMember(IsRequired = true)]
        List<SecurityCallFrame> m_StackFrames = new List<SecurityCallFrame>();

        #endregion

        #region Properties

        public SecurityCallFrame[] Calls
        {
            get
            {
                return this.m_StackFrames.ToArray();
            }
        }

        public int Count
        {
            get
            {
                return this.m_StackFrames.Count;
            }
        }

        public SecurityCallFrame LastCall
        {
            get
            {
                if (!this.m_StackFrames.Any())
                    return null;

                return this.m_StackFrames.Last();
            }
        }

        public SecurityCallFrame OriginalCall
        {
            get
            {
                if (!this.m_StackFrames.Any())
                    return null;

                return this.m_StackFrames.First();
            }
        }

        #endregion

        #region Methods

        [MethodImpl(MethodImplOptions.NoInlining)]
        public void AppendCall()
        {
#if DEBUG
            const int popIndex = 12;
#else
            const int popIndex = 11;
#endif
            this.AppendCall(popIndex);
        }

        [MethodImpl(MethodImplOptions.NoInlining)]
        internal void AppendCall(int index)
        {
            SecurityCallFrame call = new SecurityCallFrame();
            this.m_StackFrames.Add(call);
            StackFrame frame = new StackFrame(index);
						
			List<Claim> claims = new List<Claim>();
			IEnumerable<string> registeredClaimsTypes = ClaimsBuilder.RegisteredClaims();

            ClaimsIdentity currentClaimsIdentity = IdentityUserData.FindIdentity();

            if (currentClaimsIdentity != null)
            {
                foreach (string claimType in registeredClaimsTypes)
                {
                    if (currentClaimsIdentity.Claims.Any(c => c.Type == claimType))
                    {
                        claims.Add(currentClaimsIdentity.FindFirst(claimType));
                    }
                }
            }

            SetAuthenticationType(ref claims);
            SetCallerAddress(ref claims);
            SetWindowsAccountName(ref claims);
            SetFederatedAuthClaims(ref claims);
            SetCallerType(ref claims, frame.GetMethod() != null ? frame.GetMethod().DeclaringType.ToString() : string.Empty);
            this.SetActivityId(ref claims);

            call.Claims = claims;

        }
                
        internal void Clear()
        {
            this.m_StackFrames.Clear();
        }

        #endregion

        #region Claim Setting Methods

        private static void AddClaim(ref List<Claim> claims, string claimName, string value)
        {
            if (claims != null && !string.IsNullOrWhiteSpace(claimName) && !string.IsNullOrWhiteSpace(value))
                claims.Add(new Claim(claimName, value));
        }

        private static void SetActivityId(ref List<Claim> claims, string activityId)
        {
            AddClaim(ref claims, ServiceModelEx.Security.Principal.CustomClaimTypes.ActivityId, activityId);
        }

        private void SetActivityId(ref List<Claim> claims)
        {
            // Call from client
            if (this.Count == 1)
            {
                // If the claim does not already exist then add it
                var activityClaim = claims.FirstOrDefault(claim => claim.Type == ServiceModelEx.Security.Principal.CustomClaimTypes.ActivityId);
                if (activityClaim == null)
                    SetActivityId(ref claims, Guid.NewGuid().ToString());

                return;
            }

            // Call from another service
            if (this.Count > 1)
            {
                //Copy the activity ID from the previous call
                var previousCall = this.m_StackFrames[this.Count - 2];
                var activityClaim = previousCall.Claims.FirstOrDefault(claim => claim.Type == ServiceModelEx.Security.Principal.CustomClaimTypes.ActivityId);

                if (activityClaim != null)
                    SetActivityId(ref claims, activityClaim.Value);
            }
        }

        private static void SetAuthenticationType(ref List<Claim> claims)
        {
            string authenticationType = ServiceSecurityContext.Current != null ? ServiceSecurityContext.Current.PrimaryIdentity.AuthenticationType : Thread.CurrentPrincipal.Identity.AuthenticationType;
            AddClaim(ref claims, System.Security.Claims.ClaimTypes.AuthenticationMethod, authenticationType);
        }

        private static void SetCallerAddress(ref List<Claim> claims)
        {
            string address = OperationContext.Current != null && 
                OperationContext.Current.Channel != null
                && OperationContext.Current.Channel.LocalAddress != null
                && OperationContext.Current.Channel.LocalAddress.Uri != null ? OperationContext.Current.Channel.LocalAddress.Uri.ToString() : Environment.MachineName;            
            AddClaim(ref claims, ServiceModelEx.Security.Principal.CustomClaimTypes.CallerAddress, address);
        }

        private static void SetCallerType(ref List<Claim> claims, string callerType)
        {
            AddClaim(ref claims, ServiceModelEx.Security.Principal.CustomClaimTypes.CallerType, callerType);
        }

        private static void SetWindowsAccountName(ref List<Claim> claims)
        {
            string identityName = Thread.CurrentPrincipal.Identity.Name;
            if (string.IsNullOrWhiteSpace(identityName))
            {
                WindowsIdentity windowsIdentity = WindowsIdentity.GetCurrent();
                if (windowsIdentity != null)
                    identityName = windowsIdentity.Name;
            }

            AddClaim(ref claims, System.Security.Claims.ClaimTypes.WindowsAccountName, identityName);
        }

        private void SetFederatedAuthClaims(ref List<Claim> claims)
        {
            try
            {
                // We have a HttpContext check if we have any FedAuth cookies and if we do add them to our claims.
                if (HttpContext.Current != null && HttpContext.Current.Request != null && HttpContext.Current.Request.Cookies != null && HttpContext.Current.Request.Cookies.Count > 0)
                {
                    if (this.m_StackFrames.Count == 1)
                    {
                        if (HttpContext.Current.Request.Cookies["FedAuth"] != null)
                        {
                            string FedAuth = HttpContext.Current.Request.Cookies["FedAuth"].Value;
                            AddClaim(ref claims, CustomClaimTypes.FedAuth, FedAuth);
                        }

                        if (HttpContext.Current.Request.Cookies["FedAuth1"] != null)
                        {
                            string FedAuth1 = HttpContext.Current.Request.Cookies["FedAuth1"].Value;
                            AddClaim(ref claims, CustomClaimTypes.FedAuth1, FedAuth1);
                        }
                    }
                    else
                    {
                        CopyFederationClaimsFromPreviousCall(claims);
                    }
                }
                else
                {
                    CopyFederationClaimsFromPreviousCall(claims);
                }
            }
            catch (System.Web.HttpException)
            {
                // Catching and swallowing the http exception. The Http context variable is not null but unavailable when the code is being called from within the global.asax
                // There is no way to detect that the exception will occur.
                // See http://stackoverflow.com/questions/2609512/workaround-for-httpcontext-hiderequestresponse-being-internal-detect-if-httpcon
            }
        }

        private void CopyFederationClaimsFromPreviousCall(List<Claim> claims)
        {
            if (this.m_StackFrames.Count >= 2)
            {
                //Copy the activity ID from the previous call
                var previousCall = this.m_StackFrames[this.Count - 2];

                if (previousCall != null)
                {
                    Claim FedAuth = previousCall.Claims.FirstOrDefault(claim => claim.Type == CustomClaimTypes.FedAuth);
                    Claim FedAuth1 = previousCall.Claims.FirstOrDefault(claim => claim.Type == CustomClaimTypes.FedAuth1);

                    if (FedAuth != null)
                    {
                        claims.Add(FedAuth);
                    }

                    if (FedAuth1 != null)
                    {
                        claims.Add(FedAuth1);
                    }
                }
            }
        }

        #endregion
    }
}
