﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Security.Claims;
using System.Text;
using System.Threading.Tasks;

namespace ServiceModelEx.Security.Principal
{
	public static class ClaimsBuilder
	{
		public static IEnumerable<string> RegisteredClaims()
		{
			var registeredClaims = new[]
                {                    
                    ClaimTypes.NameIdentifier,
                    ClaimTypes.Name,
                    ClaimTypes.Email,
                    ClaimTypes.IsPersistent,
                    CustomClaimTypes.Culture,
                    CustomClaimTypes.Theme,
                    CustomClaimTypes.UserLanguages,
                    CustomClaimTypes.ClientWebsiteUrl,
                    CustomClaimTypes.Capabilities,
                    CustomClaimTypes.FedAuth,
                    CustomClaimTypes.FedAuth1
                };

			return registeredClaims;
		}

		public static IEnumerable<Claim> BuildClaims(
			int userId = -1,
			string userName = "",
			string email = "",
			bool rememberMe = false,
			string culture = "",
			string theme = "",
			string clientUrl = "",
			IEnumerable<int> languageIds = null,
            IEnumerable<int> capabilities = null,
            string fedAuth = "",
            string fedAuth1 = "")
		{
				
			// Only add claims if they have a value
			List<Claim> claims = new List<Claim>();

			AddClaim(ref claims, ClaimTypes.NameIdentifier, userId);
            AddClaim(ref claims, ClaimTypes.Name, userName, true);
			AddClaim(ref claims, ClaimTypes.Email, email);
			AddClaim(ref claims, ClaimTypes.IsPersistent, rememberMe);
			AddClaim(ref claims, CustomClaimTypes.Culture, culture);
			AddClaim(ref claims, CustomClaimTypes.Theme, theme);
			AddClaim(ref claims, CustomClaimTypes.ClientWebsiteUrl, clientUrl);
			AddClaim(ref claims, CustomClaimTypes.UserLanguages, languageIds);
            AddClaim(ref claims, CustomClaimTypes.Capabilities, capabilities);
            AddClaim(ref claims, CustomClaimTypes.FedAuth, fedAuth);
            AddClaim(ref claims, CustomClaimTypes.FedAuth1, fedAuth1);			

			return claims;
		}

        public static string GetClaimValue(ClaimsIdentity identity, string claimType)
        {
            if (identity != null)
            {
                Claim claim = identity.FindFirst(claimType);
                if (claim != null)
                    return claim.Value;
            }
            return string.Empty;
        }

        public static string GetClaimValue(ClaimsPrincipal principal, string claimType)
        {
            if (principal != null)
            {
                Claim claim = principal.FindFirst(claimType);
                if (claim != null)
                    return claim.Value;
            }
            return string.Empty;
        }

        public static bool GetClaimValueBool(ClaimsPrincipal principal, string claimType)
        {
            string claimValue = GetClaimValue(principal, claimType);
            if (!string.IsNullOrWhiteSpace(claimValue))
            {
                bool result;
                if (bool.TryParse(claimValue, out result))
                    return result;
            }
            return false;
        }

        public static bool GetClaimValueBool(ClaimsIdentity identity, string claimType)
        {
            string claimValue = GetClaimValue(identity, claimType);
            if (!string.IsNullOrWhiteSpace(claimValue))
            {
                bool result;
                if (bool.TryParse(claimValue, out result))
                    return result;
            }
            return false;
        }

        public static Guid GetClaimValueGuid(ClaimsIdentity identity, string claimType)
        {
            string claimValue = GetClaimValue(identity, claimType);
            if (!string.IsNullOrWhiteSpace(claimValue))
            {
                Guid result;
                if (Guid.TryParse(claimValue, out result))
                    return result;
            }
            return Guid.Empty;
        }

        public static int GetClaimValueInt(ClaimsIdentity identity, string claimType)
        {
            string claimValue = GetClaimValue(identity, claimType);
            if (!string.IsNullOrWhiteSpace(claimValue))
            {
                int value = 0;
                if (int.TryParse(claimValue, out value))
                    return value;
            }
            return 0;
        }

        public static IEnumerable<int> GetClaimValueIEnumerableInt(ClaimsIdentity identity, string claimType)
        {
            List<int> result = new List<int>();
            string claimValue = GetClaimValue(identity, claimType);
            if (!string.IsNullOrWhiteSpace(claimValue))
            {
                string[] data = claimValue.Split(',');

                foreach (string item in data)
                {
                    result.Add(Convert.ToInt32(item));
                }
            }

            return result;
        }

        private static void AddClaim(ref List<Claim> claims, string type, int value)
		{
			if (claims != null && !string.IsNullOrWhiteSpace(type) && RegisteredClaims().Contains(type))
			{
				claims.Add(new Claim(type, value.ToString()));
			}
		}

		private static void AddClaim(ref List<Claim> claims, string type, string value, bool overrideDefaultValue = false)
		{
			if (claims != null && !string.IsNullOrWhiteSpace(type) && RegisteredClaims().Contains(type) && (!string.IsNullOrWhiteSpace(value) || overrideDefaultValue))
			{
				claims.Add(new Claim(type, value));
			}
		}

		private static void AddClaim(ref List<Claim> claims, string type, bool value)
		{
			if (claims != null && !string.IsNullOrWhiteSpace(type) && RegisteredClaims().Contains(type) && value)
			{
				claims.Add(new Claim(type, value.ToString()));
			}
		}
		
		private static void AddClaim(ref List<Claim> claims, string type, IEnumerable<int> value)
		{
			if (claims != null && !string.IsNullOrWhiteSpace(type) && RegisteredClaims().Contains(type) && value != null && value.Any())
			{				
				claims.Add(new Claim(type, String.Join(",", value)));
			}
		}		
	}
}
